import { InjectionResult, PROGRAM_NAME, SaveResult, USER_AGENT, } from "./constants.js";
import { logger } from "./logger.js";
import { getRuntimeConfig } from "./runtimeConfig.js";
import { formatAsList } from "./utils.js";
export let pushNotifier;
var Event;
(function (Event) {
    Event["RESULTS"] = "RESULTS";
})(Event || (Event = {}));
export class PushNotifier {
    url;
    constructor(url) {
        this.url = url;
    }
    async notify({ title = PROGRAM_NAME, body, ...rest }) {
        if (this.url) {
            try {
                const response = await fetch(this.url, {
                    method: "POST",
                    headers: {
                        "Content-Type": "application/json",
                        "User-Agent": USER_AGENT,
                    },
                    body: JSON.stringify({ title, body, ...rest }),
                });
                if (!response.ok) {
                    logger.error({
                        message: "Failed to send push notification",
                    });
                    logger.debug({
                        message: `Server response: ${response.status} ${response.statusText}`,
                    });
                }
            }
            catch (error) {
                logger.error({
                    message: "Failed to send push notification",
                });
                logger.debug({
                    message: error,
                });
            }
        }
    }
}
export function sendResultsNotification(searchee, results) {
    const source = searchee.label;
    const notableSuccesses = results.filter(([, , actionResult]) => actionResult === InjectionResult.SUCCESS ||
        actionResult === SaveResult.SAVED);
    const failures = results.filter(([, , actionResult]) => actionResult === InjectionResult.FAILURE);
    if (notableSuccesses.length) {
        const name = notableSuccesses[0][0].metafile.name;
        const numTrackers = notableSuccesses.length;
        const infoHashes = notableSuccesses.map(([assessment]) => assessment.metafile.infoHash);
        const trackers = notableSuccesses.map(([, tracker]) => tracker);
        const trackersListStr = formatAsList(trackers, { sort: true });
        const performedAction = notableSuccesses[0][2] === InjectionResult.SUCCESS
            ? "Injected"
            : "Saved";
        pushNotifier.notify({
            body: `${source}: ${performedAction} ${name} from ${numTrackers} trackers: ${trackersListStr}`,
            extra: {
                event: Event.RESULTS,
                name,
                infoHashes,
                trackers,
                source,
                result: notableSuccesses[0][2],
            },
        });
    }
    if (failures.length) {
        const name = failures[0][0].metafile.name;
        const numTrackers = failures.length;
        const infoHashes = failures.map(([assessment]) => assessment.metafile.infoHash);
        const trackers = failures.map(([, tracker]) => tracker);
        const trackersListStr = formatAsList(trackers, { sort: true });
        pushNotifier.notify({
            body: `Failed to inject ${name} from ${numTrackers} trackers: ${trackersListStr}`,
            extra: {
                event: Event.RESULTS,
                name,
                infoHashes,
                trackers,
                source,
                result: failures[0][2],
            },
        });
    }
}
export function initializePushNotifier() {
    const { notificationWebhookUrl } = getRuntimeConfig();
    pushNotifier = new PushNotifier(notificationWebhookUrl);
}
export function sendTestNotification() {
    pushNotifier.notify({ body: "Test", extra: { event: "TEST" } });
    logger.info("Sent test notification");
}
//# sourceMappingURL=pushNotifier.js.map